window.Gleam.initSearch([{"doc":"chip","title":"chip","content":"# Chip - A subject registry for Gleam\n\n[![Package Version](https://img.shields.io/hexpm/v/chip)](https://hex.pm/packages/chip)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/chip/)\n\nChip is a performant local registry that can hold to a set of [subjects](https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject) individually or as part of a group. \n\n## Example\n\nOne of the most useful use cases for chip is broadcasting messages to registered subjects:\n\n```gleam\nimport artifacts/game.{DrawCard, FireDice, PlayChip}\nimport chip\n\npub fn main() {\n  let assert Ok(registry) = chip.start()\n\n  let assert Ok(session_a) = game.start(DrawCard)\n  let assert Ok(session_b) = game.start(FireDice)\n  let assert Ok(session_c) = game.start(PlayChip)\n\n  chip.register(registry, chip.new(session_a))\n  chip.register(registry, chip.new(session_b))\n  chip.register(registry, chip.new(session_c))\n\n  chip.dispatch(registry, fn(session) {\n    game.next(session)\n  })\n}\n```\n\n## Features\n\nChip was designed with a very minimal but practical feature set:\n\n* Subjects may be individually retrieved via tags.\n* It is also possible to dispatch actions to groups of Subjects.\n* Chip will automatically delist dead processes.\n\nFor more possible use-cases check the documented guidelines.\n\n## Development \n\nFrom now on updates will focus on reliability and performance, but new additions to the API will be considered with care. Features are [documented as Issues](https://github.com/chouzar/chip/issues?q=is%3Aopen+is%3Aissue+label%3Aenhancement) on the project&#39;s repo, if you have questions or like to se a new feature please open an issue.\n\n### Previous Art \n\nThis registry takes and combines some ideas from:\n\n* Elixir’s [registry](https://hexdocs.pm/elixir/Kernel.html) module.\n* Erlang&#39;s [pg](https://www.erlang.org/doc/apps/kernel/pg.html) module.\n* The [syn](https://github.com/ostinelli/syn) global registry library.\n\n### Other Gleam registry libraries\n\nOther registry libraries will provide different semantics and functionality:\n\n* [Singularity](https://hexdocs.pm/singularity/) is designed to register a fixed number of actors, each of which may have a different message type.\n\n## Installation\n\n```sh\ngleam add chip\n```\n","url":"index.html"},{"doc":"chip","title":"chip","content":"# Chip as your local PubSub system\n\nA PubSub will allow us to &quot;subscribe&quot; subjects to a &quot;topic&quot;, then later we we can &quot;publish&quot; an event to all subscribed subjects. \n\n```gleam\nimport chip\nimport gleam/erlang/process\nimport gleam/otp/supervisor\n\npub type PubSub(message, channel) =\n  chip.Registry(message, Nil, channel)\n\npub fn start() {\n  chip.start()\n}\n\npub fn childspec() {\n  supervisor.worker(fn(_param) { start() })\n  |&gt; supervisor.returning(fn(_param, pubsub) { pubsub })\n}\n\npub fn subscribe(\n  pubsub: PubSub(message, channel),\n  channel: channel,\n  subject: process.Subject(message),\n) -&gt; Nil {\n  chip.new(subject)\n  |&gt; chip.group(channel)\n  |&gt; chip.register(pubsub, _)\n}\n\npub fn publish(\n  pubsub: PubSub(message, channel),\n  channel: channel,\n  message: message,\n) -&gt; Nil {\n  chip.dispatch_group(pubsub, channel, fn(subscriber) {\n    process.send(subscriber, message)\n  })\n}\n```\n\nThe pattern above would be a generic way of re-defining chip as a pubsub system, we may use it to wire-up applications that require reacting to events. For example, lets assume we want to create a chat application.  \n\n```gleam \nimport artifacts/pubsub\nimport chip\nimport gleam/erlang/process\nimport gleam/int\nimport gleam/list\nimport gleam/otp/task\n\npub type Channel {\n  General\n  Coffee\n  Pets\n}\n\npub type Event {\n  Event(id: Int, message: String)\n}\n\npub fn main() {\n  let assert Ok(pubsub) = pubsub.start()\n\n  // for this scenario, out of simplicity, all clients are the current process.\n  let client = process.new_subject()\n\n  // client is interested in coffee and pets\n  chip.register(pubsub, chip.new(client) |&gt; chip.group(Coffee))\n  chip.register(pubsub, chip.new(client) |&gt; chip.group(Pets))\n\n  // lets assume this is the server process broadcasting a welcome message\n  task.async(fn() {\n    chip.dispatch_group(pubsub, General, fn(client) {\n      Event(id: 1, message: &quot;Welcome to General! Follow rules and be nice.&quot;)\n      |&gt; process.send(client, _)\n    })\n    chip.dispatch_group(pubsub, Coffee, fn(client) {\n      Event(id: 2, message: &quot;Ice breaker! Favorite cup of coffee?&quot;)\n      |&gt; process.send(client, _)\n    })\n    chip.dispatch_group(pubsub, Pets, fn(client) {\n      Event(id: 3, message: &quot;Pets!&quot;)\n      |&gt; process.send(client, _)\n    })\n  })\n\n  // it is then each client&#39;s responsability to listen to incoming messages\n\n  // client only receives coffee and pets messages\n  let assert True =\n    listen_for_messages(client, [])\n    |&gt; list.all(fn(message) {\n      case message {\n        &quot;2: Ice breaker! Favorite cup of coffee?&quot; -&gt; True\n        &quot;3: Pets!&quot; -&gt; True\n        _other -&gt; False\n      }\n    })\n}\n\nfn listen_for_messages(client, messages) -&gt; List(String) {\n  // this function will listen until messages stop arriving for 100 milliseconds\n\n  // a selector is useful to transform our Events into types a client expects (String).\n  let selector =\n    process.new_selector()\n    |&gt; process.selecting(client, to_string)\n\n  case process.select(selector, 100) {\n    Ok(message) -&gt;\n      // a message was received, capture it and attempt to listen for another message\n      message\n      |&gt; list.prepend(messages, _)\n      |&gt; listen_for_messages(client, _)\n\n    Error(Nil) -&gt;\n      // a message was not received, stop listening and return captured messages in order\n      messages\n      |&gt; list.reverse()\n  }\n}\n\nfn to_string(event: Event) -&gt; String {\n  int.to_string(event.id) &lt;&gt; &quot;: &quot; &lt;&gt; event.message\n}\n```\n\nThere are many ways to structure a PubSub system in Gleam so this guide is just a starting point. \n\nWhile building your system with a PubSub you may start getting into clashes between modules (gleam doesn&#39;t like circular dependencies), if you find yourself in this situation try to:\n\n* Divide shared types, for example put `Event` in its own module.  \n* Take advantage of generics so you don&#39;t have to be bound to specific types.\n* Take advantage of callbacks so you don&#39;t have to be bound to specific behaviour.\n\nAs a final tip, sometimes we may try to optimize the project structure too early, if in doubt try to keep your types and functions in one module and separate as your abstractions and domain knowledge mature.\n","url":"guides/as-local-pubsub.html"},{"doc":"chip","title":"chip","content":"# Chip as a session index\n\nThe main purpose of the process index is to be able to &quot;track&quot; individual subjects through an identifier in your application. The identifier may be an integer, string or even a union type if you only need few records.\n\n```gleam\nimport chip\nimport gleam/erlang/process\nimport gleam/otp/actor\nimport gleam/otp/supervisor\n\npub type Store(message) =\n  chip.Registry(message, Int, Nil)\n\npub type Id =\n  Int\n\npub fn start() -&gt; Result(Store(message), actor.StartError) {\n  chip.start()\n}\n\npub fn childspec() {\n  supervisor.worker(fn(_index) { start() })\n}\n\npub fn index(store: Store(message), id: Id, subject: process.Subject(message)) {\n  chip.new(subject)\n  |&gt; chip.tag(id)\n  |&gt; chip.register(store, _)\n}\n\npub fn get(store: Store(message), id: Id) {\n  chip.find(store, id)\n}\n```\n\nProbably one of the most common uses cases for chip will involve looking up for subjects in your system. \n\nFor example, in a web app we may add the registry as part of our app context. \n\n```gleam\nimport artifacts/game.{DrawCard}\nimport artifacts/store\n\npub fn main() {\n  let assert Ok(sessions) = store.start()\n\n  let assert Ok(session_1) = game.start(DrawCard)\n  let assert Ok(session_2) = game.start(DrawCard)\n  let assert Ok(session_3) = game.start(DrawCard)\n\n  store.index(sessions, 1, session_1)\n  store.index(sessions, 2, session_2)\n  store.index(sessions, 3, session_3)\n\n  router(sessions, &quot;/resource/&quot;, 2)\n}\n\nfn router(sessions, url, id) {\n  case url, id {\n    &quot;/resource/&quot;, id -&gt; render_resource(sessions, id)\n    _other, _id -&gt; render_error()\n  }\n}\n\nfn render_resource(sessions, id) {\n  case store.get(sessions, id) {\n    Ok(session) -&gt; game.current(session)\n    Error(Nil) -&gt; render_error()\n  }\n}\n\nfn render_error() {\n  &quot;Print Error&quot;\n}\n```\n\nOf course, this ability to directly reference subjects is not very useful without a supervision tree, because if the subject dies we can no longer send messages to it. Check the [supervision guideline](as-part-of-supervisor.html) for more. \n","url":"guides/as-subject-index.html"},{"doc":"chip","title":"chip","content":"# Chip as part of a Supervision tree\n\nA supervision tree is a strategy used in the wider erlang ecosystem to keep long-running processeess alive. When a process in the tree terminates the supervisor will attempt to restart the process and any subsequent processess in the tree, giving our whole system self-healing capabilities.\n\nTo make chip and other subjects part of a supervision tree we first need to define their respective child specifications, these specifications define their behaviour and state when starting or re-starting.\n\n```gleam\nimport artifacts/game\nimport chip\nimport gleam/erlang/process\nimport gleam/otp/supervisor\n\ntype Registry =\n  chip.Registry(game.Message, Int, Nil)\n\npub type Context {\n  Context(caller: process.Subject(Registry), registry: Registry, id: Int)\n}\n\npub fn registry() {\n  // The registry childspec first starts the registry.\n  supervisor.worker(fn(_caller: process.Subject(Registry)) { chip.start() })\n  // After starting we transform the parameter from caller into a context for \n  // the sessions we want to register. \n  |&gt; supervisor.returning(fn(caller, registry) { Context(caller, registry, 1) })\n}\n\npub fn game() {\n  supervisor.worker(fn(context: Context) {\n    game.start_with(context.registry, context.id, game.DrawCard)\n  })\n  |&gt; supervisor.returning(fn(context: Context, _game_session) {\n    Context(..context, id: context.id + 1)\n  })\n}\n\npub fn ready() {\n  // This childspec is a noop addition to the supervisor, on return it\n  // will send back the registry reference.\n  supervisor.worker(fn(_context: Context) { Ok(process.new_subject()) })\n  |&gt; supervisor.returning(fn(context: Context, _self) {\n    process.send(context.caller, context.registry)\n    Nil\n  })\n}\n```\n\nThen integrating our specs to a supervisor is simple enough: \n\n```gleam \nimport artifacts/spec\nimport chip\nimport gleam/erlang/process\nimport gleam/list\nimport gleam/otp/supervisor\n\npub fn main() {\n  let self = process.new_subject()\n\n  let assert Ok(_supervisor) =\n    supervisor.start_spec(\n      supervisor.Spec(\n        argument: self,\n        max_frequency: 5,\n        frequency_period: 1,\n        init: fn(children) {\n          children\n          // First spawn the registry\n          |&gt; supervisor.add(spec.registry())\n          // Then spawn 1 to 50 game sessions\n          |&gt; list.fold(\n            list.range(1, 50),\n            _,\n            fn(children, _id) { supervisor.add(children, spec.game()) },\n          )\n          // Finally notify the main process we&#39;re ready\n          |&gt; supervisor.add(spec.ready())\n        },\n      ),\n    )\n\n  let assert Ok(registry) = process.receive(self, 500)\n  let assert Ok(_session) = chip.find(registry, 33)\n}\n```\n\nWith the implementation above, we have asured that we will have our 1 to 50 game sessions always available for retireval.  \n","url":"guides/as-part-of-supervisor.html"},{"doc":"chip","title":"Chip","content":"pub opaque type Chip(msg, tag, group)\n A &quot;chip&quot; used for registration. Check the [new](#new) function.\n\nSynonyms:\nchip.Chip\nchip Chip","url":"chip.html#Chip"},{"doc":"chip","title":"Message","content":"pub opaque type Message(msg, tag, group)\n Chip&#39;s internal message type.\n\nSynonyms:\nchip.Message\nchip Message","url":"chip.html#Message"},{"doc":"chip","title":"Registry","content":"pub type Registry(msg, tag, group) =\n  process.Subject(Message(msg, tag, group))\n An shorter alias for the registry&#39;s Subject. \n \n Sometimes, when building out your system it may be useful to state the Registry&#39;s types.\n \n ## Example\n \n \n let assert Ok(registry) = chip.start()\n let registry: chip.Registry(Event, Id, Topic)\n \n \n Which is equivalent to: \n \n \n let assert Ok(registry) = chip.start()\n let registry: process.Subject(chip.Message(Event, Id, Topic))\n \n \n By specifying the types we can document the kind of registry we are working with.\n\nSynonyms:\nchip.Registry\nchip Registry","url":"chip.html#Registry"},{"doc":"chip","title":"dispatch","content":"pub fn dispatch(\n  registry: Subject(Message(a, b, c)),\n  callback: fn(Subject(a)) -&gt; Nil,\n) -&gt; Nil\n Applies a callback over all registered Subjects.\n \n ## Example\n \n \n chip.dispatch(registry, fn(subject) { \n   process.send(subject, message)\n })\n \nSynonyms:\nchip.dispatch\nchip dispatch","url":"chip.html#dispatch"},{"doc":"chip","title":"dispatch_group","content":"pub fn dispatch_group(\n  registry: Subject(Message(a, b, c)),\n  group: c,\n  callback: fn(Subject(a)) -&gt; Nil,\n) -&gt; Nil\n Applies a callback over a group.\n \n ## Example\n \n \n chip.dispatch_group(registry, Pets, fn(subject) { \n   process.send(subject, message)\n })\n \nSynonyms:\nchip.dispatch_group\nchip dispatch_group","url":"chip.html#dispatch_group"},{"doc":"chip","title":"find","content":"pub fn find(\n  registry: Subject(Message(a, b, c)),\n  tag: d,\n) -&gt; Result(Subject(a), Nil)\n Retrieves a tagged subject.\n \n ## Example\n \n \n let assert Ok(subject) = chip.find(registry, &quot;Luis&quot;)\n \nSynonyms:\nchip.find\nchip find","url":"chip.html#find"},{"doc":"chip","title":"group","content":"pub fn group(\n  registrant: Chip(a, b, c),\n  group: c,\n) -&gt; Chip(a, b, c)\n Adds the &quot;chip&quot; under a group. \n \n ## Example\n \n \n chip.new(subject)\n |&gt; chip.group(General) \n \nSynonyms:\nchip.group\nchip group","url":"chip.html#group"},{"doc":"chip","title":"new","content":"pub fn new(subject: Subject(a)) -&gt; Chip(a, b, c)\n Creates a new &quot;chip&quot; that can be tagged, grouped and registered. \n \n ## Example\n \n \n chip.new(subject)\n \nSynonyms:\nchip.new\nchip new","url":"chip.html#new"},{"doc":"chip","title":"register","content":"pub fn register(\n  registry: Subject(Message(a, b, c)),\n  registrant: Chip(a, b, c),\n) -&gt; Nil\n Registers a &quot;chip&quot;. \n \n ## Example\n \n \n let assert Ok(registry) = chip.start()\n \n chip.new(subject)\n |&gt; chip.register(registry, _)\n \n \n The subject may be registered under a tag or group.\n \n \n let assert Ok(registry) = chip.start()\n \n chip.new(subject)\n |&gt; chip.tag(&quot;Francisco&quot;) \n |&gt; chip.group(Coffee) \n |&gt; chip.register(registry, _)\n \n \n You may register any subject at any point in time but usually keeping it under the initialization\n step of your process (like an Actor&#39;s `init` callback) will keep things organized and tidy. \nSynonyms:\nchip.register\nchip register","url":"chip.html#register"},{"doc":"chip","title":"start","content":"pub fn start() -&gt; Result(Subject(Message(a, b, c)), StartError)\n Starts the registry.\n \n ## Example\n \n \n &gt; chip.start()\n \nSynonyms:\nchip.start\nchip start","url":"chip.html#start"},{"doc":"chip","title":"stop","content":"pub fn stop(registry: Subject(Message(a, b, c))) -&gt; Nil\n Stops the registry.\n \n ## Example\n \n \n let assert Ok(registry) = chip.start()\n chip.stop(registry)\n \nSynonyms:\nchip.stop\nchip stop","url":"chip.html#stop"},{"doc":"chip","title":"tag","content":"pub fn tag(registrant: Chip(a, b, c), tag: b) -&gt; Chip(a, b, c)\n Adds a unique tag to a &quot;chip&quot;, it will overwrite any previous subject under the same tag.\n \n ## Example\n \n \n chip.new(subject)\n |&gt; chip.tag(&quot;Luis&quot;) \n \nSynonyms:\nchip.tag\nchip tag","url":"chip.html#tag"},{"doc":"chip","title":"chip","content":" Chip is a local [subject](https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject) \n registry that can reference subjects individually or as part of a group. Will also \n automatically delist dead processes.","url":"chip.html"}]);